<?php
/******************************************************************************************************
 * Project: PHP IMSP CLIENT LIBRARY 
 * File: cIMSPABook.PHP
 * Purpose: This file provides the means to work with IMSP ADDRESSBOOKS
 *
 * Copyright 2002 Michael Rubinsky <mike@theupstairsroom.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * The file "IMSPStream.PHP" must be present in the same directory for this to work
 * since this class extends the IMSPStream Class
 *
 *
 ********************************************************************************************************
 *PROTOCOL IMPLEMENTATION STATUS:
 *      GENERAL
 *              -LOGIN-----------------------------DONE 
 *                              supports:
 *                                      -LOGIN (plaintext)
 *                                      -AUTHENTICATE CRAM-MD5
 *              -LOGOUT----------------------------DONE                                                                 
 *      ADDRESSBOOKS
 *              -ADDRESSBOOK-----------------------DONE
 *              -CREATEADDRESSBOOK-----------------DONE
 *              -DELETEADDRESSBOOK-----------------DONE
 *              -RENAMEADDRESSBOOK-----------------DONE
 *              -SEARCHADDRES----------------------DONE
 *              -FETCHADDRESS----------------------DONE
 *              -STOREADDRESS----------------------DONE
 *              -DELETEADDRESS---------------------DONE
 *      LOCKING
 *              -LOCKADDRESSBOOK-------------------DONE
 *              -UNLOCKADDRESSBOOK-----------------DONE
 *      ACL
 *              -SETACL ADDRESSBOOK----------------DONE
 *              -DELETEACL ADDRESSBOOK-------------DONE
 *              -GETACL ADDRESSBOOK----------------DONE
 *              -MYRIGHTS ADDRESSBOOK--------------DONE
 *******************************************************************************************************
 *Example of use:
 *
 *      $myABook = new IMSPAddressBook([imsp_server_name],[port]);      
 *      $myABook->logging = TRUE;                                                                               //Turn logging on or off (on by default)
 *  $myABook->log_level = 1;                                                                            //Level 1(default)= normal Level 2 = Verbose
 *  $myABook->log_name = "path_to_log_file"     ;                                               //directory will have to be writable by the web server (nobody)
 *  $myABook->logon("username","password");                                             
 *      
 *  $entry = $myABook->getAddressBookEntry("abookName","entryName"); 
 *      .
 *  .
 *  .
 *  $myABook->logout();
 *
 ******************************************************************************************************
 *See the sample files for examples on how to use this class or look at the comments in the function
 *headers for more information.
 *****************************************************************************************************/
 
 //Include the parent Class description
 require("cIMSPStream2.php");

//These define regExp that should match the respective server response strings
define ("IMSP_ADDRESSBOOK_RESPONSE","/^\* ADDRESSBOOK/");
define ("IMSP_SEARCHADDRESS_RESPONSE","/^\* SEARCHADDRESS/");
define ("IMSP_FETCHADDRESS_RESPONSE","/^\* FETCHADDRESS /");
define ("IMSP_ACL_RESPONSE","/^\* ACL ADDRESSBOOK/");
define ("IMSP_MYRIGHTS_RESPONSE","/^\* MYRIGHTS ADDRESSBOOK/");
define ("IMSP_OCTET_COUNT","/({)([0-9]{1,})(\}$)/");
define ("IMSP_MUST_USE_LITERAL","/[^a-z0-9\s\-,]/i");

//Addressbook specific exit codes
define ("IMSP_ENTRY_LOCKED","That addressbook entry is locked or cannot be unlocked.");
define ("IMSP_ENTRY_NOT_FOUND","No entry in this addressbook matches your query.");

//string of supported ACL rights
define ("IMSP_ACL_RIGHTS","lrwcda");

                                                                        
class IMSPAddressBook extends IMSPStream {
 //Assign some properties that pertain to addresbooks only.

//Set this to "ascend" or "descend" to determine the order the names are returned in.
 var $sort = "ascend";  

 // Method Definitions
 /***************************************************************************************
  *Contructor Function
 /***************************************************************************************/
 function IMSPAddressBook($local_imsp_server="", $local_port="") {
        //Simply delegate this method to the parent.
        if (!$this->IMSPStream($local_imsp_server,$local_port)) {
                //login failure...the parent class will handle these errors.
                return FALSE;
        } else {                                
                return TRUE;
        }
}       
        
/****************************************************************************************
 *getAddressBookList() - Returns an array containing the names of all the addressbooks
 *available to the logged in user.
 *****************************************************************************************/
function getAddressBookList() {
        
                $command_string = "ADDRESSBOOK *";
                
                if (!$this->imspSend($command_string)) {
                                $this->exitCode = IMSP_CONNECTION_FAILURE;
                                $this->imspError($command_string);
                                return FALSE;   
                } 
                        
                        //iterate through the response and populate an array of addressbook names
                        $server_response = $this->imspRecieve();
                        
                        while (preg_match(IMSP_ADDRESSBOOK_RESPONSE,$server_response)){
                                /*****************************************************************
                                 *If this is a ADDRESSBOOK response, then this will split as so:
                                 * [0] and [1] can be discarded
                                 * [2] = attributes
                                 * [3] = delimiter
                                 * [4] = addressbook name
                                 ****************************************************************/
                                
                                //First, check for a {}
                                if (preg_match(IMSP_OCTET_COUNT,$server_response,$tempArray)){
                                        $abooks[] = $this->recieveStringLiteral($tempArray[2]);

                                        //Get the CRLF at end of ADDRESSBOOK response that the {} does not include.
                                        $this->recieveStringLiteral(2);
                                } else {
                                        $entry = split(" ",$server_response);
                                        $abooks[] = $entry[4];          
                                }
                                $server_response = $this->imspRecieve();        
                        }
                        
                        //Success?
                        if ($server_response != "OK") {
                                $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                                $this->imspError("Error in getAddressBookList() call");
                                return FALSE;
                        }
                        
                        
                        if ($this->log_stream) {
                                $this->write_to_log("ADDRESSBOOK command OK.");
                        }
                        return $abooks;         
}

/***************************************************************
 * getAddressBookListXML()
 * Returns a XML string containing a list of available 
 * addressbooks along with the ACL for the current user
 * as a attribute called "ACL"
 ***************************************************************/
 function getAddressBookListXML() {
        
        //First, get the information
        $list = $this->getAddressBookList();
        
        //Now build the XML string
        $xml = IMSP_XML_HEAD;
        
        $xml = "<addressbooks>\n";
        $count = count($list);
        
        for ($i=0 ; $i < $count ; $i++) {
                $abook = $list[$i];                                                     
                
                //Get the acl for this addressbook for the currently logged in user.
                $acl = $this->myRights($abook); 
                $xml .= "\t<addressbook acl='" . $acl . "'>" . $list[$i] . "</addressbook>\n";
        }
        
        $xml .= "</addressbooks>\n";
        
        return $xml;
}

/****************************************************************************************
 *searchAddressBook() - returns an array containing the names that
 *match $search critera in the addressbook named $abook
 *
 *If the criterea does not produce a match, then FALSE is returned.
 ***************************************************************************************/
function searchAddressBook($abook,$search,$searchtype="name"){
                
                if (!$this->imspSend("SEARCHADDRESS ",TRUE,FALSE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }

                //Do we need to send the abook name as {}
                if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                        $biBook = sprintf("{%d}",strlen($abook));

                        if(!$this->imspSend($biBook,FALSE,TRUE)) {
                                //Error
                                $this->exitCode = IMSP_CONNECTION_FAILURE;
                                $this->imspError();
                                return FALSE;
                        }
                        
                        if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                                $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                                $this->imspError(IMSP_NO_CONTINUATION_RESPONSE);
                                return FALSE;
                        }
                }

                
                if (!$this->imspSend("$abook $searchtype ",FALSE,FALSE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }


                //How about the search term as a {}
                if (preg_match(IMSP_MUST_USE_LITERAL,$search)){
                        $biSearch = sprintf("{%d}",strlen($search));
                        
                        if (!$this->imspSend($biSearch,FALSE,TRUE)) {
                                $this->exitCode = IMSP_CONNECTION_FAILURE;
                                $this->imspError();
                                return FALSE;
                        }


                        if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                                $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                                $this->imspError(IMSP_NO_CONTINUATION_RESPONSE);
                                return FALSE;
                        }

                        if (!$this->imspSend($search,FALSE,TRUE)) {
                                $this->exitCode = IMSP_CONNECTION_FAILURE;
                                $this->imspError();
                                return FALSE;
                        }
                
                } else {
                        
                        if (!$this->imspSend("\"$search\"",FALSE,TRUE)) {
                                $this->exitCode = IMSP_CONNECTION_FAILURE;
                                $this->imspError();
                                return FALSE;
                        }
                }
                        
        
                $server_response = $this->imspRecieve();
                
                while (preg_match(IMSP_SEARCHADDRESS_RESPONSE,$server_response)){
                        $chopped_response = preg_replace(IMSP_SEARCHADDRESS_RESPONSE,"",$server_response);

                        //Get rid of any lingering quotes.
                        $chopped_response = preg_replace("/\"/","",$chopped_response);  

                        //Remove any lingering white space in front only.
                        $temp = ltrim($chopped_response);                                                                                                               
                        
                        //vegstart
                        if(preg_match("/({)([0-9]{1,})(\}$)/",$temp,$tempArray)) {
                                $dataSize = $tempArray[2];
                $temp = $this->recieveStringLiteral($dataSize);
                                                        
                                //CRLF..since each SEARCHADDRESS response ends with a CRLF and the {} does not include it
                $this->recieveStringLiteral(2);
                        }
                        //vegend
                        
                        if($this->escape_special_characters) {
                                $temp = htmlentities($temp);
                        }
                        
                        $abookNames[] = $temp;                                              
                        
                        //Get the next response line from the server.
                        $server_response = $this->imspRecieve();                                                        
                }
                
                //Should check for OK or BAD here just to be certain
                switch($server_response) {
                
                        //This is a problem with the logic in the code so it should throw an error
                        case "BAD":                                                                                                     
                                $this->exitCode = IMSP_SYNTAX_ERROR;
                                $this->imspError($command_text);
                                return FALSE;
                        
                                
                        //Note that Cyrus IMSP wont return a NO in this case, it simply returns a OK
                        //Without any SEARCHADDRESS response lines
                        case "NO":
                                $this->exitCode = IMSP_ENTRY_NOT_FOUND;
                                return FALSE;
                }
                
                //This allows for Cyrus IMSP server not returning a "NO", only a "OK" without any "SEARCHADDRESS" response lines.
                if (count($abookNames) < 1) {
                        $this->exitCode = IMSP_ENTRY_NOT_FOUND;
                        return FALSE;
                }
                                
                if($this->log_stream) {
                        $this->write_to_log("SEARCHADDRESS command OK");
                }
                
                //Determine the sort direction and perform the sort.
                switch ($this->sort) {  
                        case "ascend":
                                sort($abookNames);
                                break;
                        case "descend":
                                rsort($abookNames);
                                break;
                }       
                                                
                return $abookNames;
}


/*****************************************************************
 * Returns an XML instance containing the names off all entries
 * that match the search critera in the specified addressbook
 *****************************************************************/
function searchAddressBookXML ($abook,$search) {

        //First get the information from our existing function
        $results = $this->searchAddressBook($abook,$search);
        
        //Now create the XML instance
        $xml = IMSP_XML_HEAD;
        $xml .= "<entrynames>\n";
        
        //Protects against zero results
        if (!$results) {
                $count = 0;
        } else {
                $count = count($results);
        }
        
        
        for ($i = 0 ; $i < $count ; $i++) {
                $xml .= "\t<entryname>" . $results[$i] . "</entryname>\n";
        }
        
        $xml .= "</entrynames>";
        
        return $xml;
}
        
        
        
        
        
                
/****************************************************************************************************
 * getAddressBookEntry() - Returns an associative array containing key - values pairs that corespond
 * to the addressbook fields and values.  Note that there will always be a "name" entry and also note
 * that the resulting values may need to be escaped for display in a web browser becuase they may 
 * contain email addresses in the form :: myName <myname@somewhere.com> 
 *
 * If no matches are found, or there is an error, a FALSE is returned.
 *****************************************************************************************************/

function getAddressBookEntry($abook,$entryName){
        //Start sending
        
        if (!$this->imspSend("FETCHADDRESS ",TRUE,FALSE)) {
                $this->exitCode = IMSP_CONNECTION_FAILURE;
                $this->imspError();
                return FALSE;
        }
        
        
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }
                
                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                        //Error
                        $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                        $this->imspError(IMSP_NO_CONTINUATION_RESPONSE);
                        return FALSE;
                }

        }

        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }
        
        if (preg_match(IMSP_MUST_USE_LITERAL,$entryName)) {
                $biName = sprintf("{%d}",strlen($entryName));

                if (!$this->imspSend($biName,FALSE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                        //Error
                        $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                        $this->imspError(IMSP_NO_CONTINUATION_RESPONSE);
                        return FALSE;

                }
                
                if (!$this->imspSend($entryName,FALSE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }

        } else {
                
                if (!$this->imspSend("\"$entryName\"",FALSE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }
        
        }
        
        $server_response = $this->imspRecieve();                                                                                                

        //Should allow for a "BAD" or "NO" response 
        switch($server_response) {
                case "BAD":
                        $this->exitCode = IMSP_SYNTAX_ERROR;
                        $this->imspError();
                        return FALSE;
                case "NO":
                        $this->exitCode = IMSP_ENTRY_NOT_FOUND;
                        //Is this really a "error"?
                        //$this->imspError("Could not find $name in $abook");                                                   
                        return FALSE;
        }

        //Get the data in an associative array
        $entry = $this->parseFetchAddressResponse($server_response);                                                    
        
        //Get the next server response -- this *should* be the OK response.
        $server_response = $this->imspRecieve();
        
        if($server_response != "OK"){
                //Unexpected response
                $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                $this->imspError();
        }
        
        if ($this->log_stream) {
                $this->write_to_log("FETCHADDRESS completed OK");
        }
        return $entry;
                                
}

function getAddressBookEntryXML($abook,$name){

        //First get the information
        $results = $this->getAddressBookEntry($abook,$name);
        
        //Build the XML instance.
        $xml = IMSP_XML_HEAD;
        
        $xml .= $this->getXMLAddressbookEntry($results);
        
        return $xml;
}

/****************************************************************************************************
 * Returns an XML instance containing all the full entries contained in the addressbook.
 * TODO: Better error handling
/****************************************************************************************************/

function getFullAddressBookXML($abook) {

        //First get the list of names in the addressbook
        $names = $this->searchAddressBook($abook,"*");
        
        if (!$names) {
                //something really wrong here!
                //Maybe no addressbook by this name?
                exit;
        }
        
        //Start the XML instance        
        $xml = IMSP_XML_HEAD;
        $xml .= "<addressbook>\n";
        
        
        //Now build the addressbookentry section
        $count = count($names);
        
        //Loop through the list of names in this addressbook, and create an XML string for that entry.
        for($i=0 ; $i < $count ; $i++) {
                $entry = $this->getAddressBookEntry($abook,$names[$i]);
                $xml .= $this->getXMLAddressbookEntry($entry);
        }
        
        $xml .= "</addressbook>\n";
        return $xml;
}
                        
        
        
/****************************************************************************************************
* Creates a new addressbook.  Takes the name of the new book. Note that this should be the 
* FULLY QUALIFIED heirarchy such as "jdoe.public" or "jdoe.clients" etc...
* Returns TRUE of FALSE
/****************************************************************************************************/
function createAddressBook($abookName){
        $command_text = "CREATEADDRESSBOOK ";                                                           

        if (preg_match(IMSP_MUST_USE_LITERAL,$abookName)) {
                $biBook = sprintf("{%d}",strlen($abookName));
                
                if (!$this->imspSend($command_text . $biBook,TRUE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }

        } else { 
                
                if (!$this->imspSend($command_text . $abookName,TRUE,TRUE)) {
                        $this->exitCode = IMSP_CONNECTION_FAILURE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if ($biBook) {
                        if(!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                                $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                                $this->imspError(IMSP_NO_COMMAND_CONTINUATION_RESPONSE);
                                return FALSE;
                        } else {
                        
                                $this->imspSend($abookName,FALSE,TRUE);
                        }
        }

        
        $server_response = $this->imspRecieve();                                                                        
        
        if(!$server_response){                                                                                                          
                //response did not come
                $this->imspError($command_text);
                return FALSE;
        }
        
        switch ($server_response) {
                case "OK":
                        $this->exitCode = IMSP_EXIT_OK;
                
                        if ($this->log_stream) {
                                $this->write_to_log("CREATEADDRESSBOOK completed OK");
                        }

                        return TRUE;                                                                                                                    
                
                case "NO":
                        //Could not create abook
                        $this->exitCode = IMSP_EXIT_BAD_MAILBOX_NAME;
                        $this->imspError();
                        return FALSE;
                
                case "BAD":
                        $this->imspError();
                        return FALSE;
                
                
                default:
                        //something unexpected!
                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                        $this->imspError($server_response);
                        return FALSE;
        }
                
}

/****************************************************************************************************
* Deletes an addressbook completely!  Returns true or false.
* TODO: Error handling for inspSend() calls.
/****************************************************************************************************/
function deleteAddressBook($abookName) {
        
        $command_text = "DELETEADDRESSBOOK " ;
        
        //Check need for {}
        if (preg_match(IMSP_MUST_USE_LITERAL,$abookName)) {
                $biBook = sprintf("{%d}",strlen($abookName));

                if (!$this->imspSend($command_text . $biBook,TRUE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

        } else { 
                
                if (!$this->imspSend($command_text . $abookName,TRUE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

        }

        if ($biBook) {
                if(!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                        $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                        $this->imspError(IMSP_NO_COMMAND_CONTINUATION_RESPONSE);
                        return FALSE;
                } else {

                        if (!$this->imspSend($abookName,FALSE,TRUE)) {
                                $this->imspError();
                                return FALSE;
                        }

                }


        }
        
        $server_response = $this->imspRecieve();                                                                        
        
        if(!$server_response){                                                                                                          
                //response did not come
                $this->imspError($command_text);
                return FALSE;
        }
        
        switch ($server_response) {
                case "OK":
                        $this->exitCode = IMSP_EXIT_OK;
                
                        if ($this->log_stream) {
                                $this->write_to_log("DELETEADDRESSBOOK completed OK");
                        }

                        return TRUE;
                
                case "NO":
                        //Could not DELETE abook
                        $this->exitCode = IMSP_EXIT_BAD_MAILBOX_NAME;
                        $this->imspError("Addressbook name: $abookName");
                        return FALSE;
                
                case "BAD":
                        $this->imspError();
                        return FALSE;
                
                
                default:
                        //something unexpected!
                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                        $this->imspError($server_response);
                        return FALSE;
        }
}
/****************************************************************************************************
* Renames an addressbook.  Takes the Oldname and the NewName 
* Returns true or false.
* TODO:  Error handling for imspSend() calls.
/****************************************************************************************************/
function renameAddressBook($abookOldName, $abookNewName) {

        //build the command while we send it...
        if (!$this->imspSend("RENAMEADDRESSBOOK ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        

        if (preg_match(IMSP_MUST_USE_LITERAL,$abookOldName)) {
                $biOldName = sprintf("{%d}",strlen($abookOldName));
                
                if (!$this->imspSend($biOldName,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                $this->imspRecieve();   //OK+
        }
                
        if (!$this->imspSend("$abookOldName ",FALSE,FALSE)) {
                $this-imspError();
                return FALSE;
        }
        

        if (preg_match(IMSP_MUST_USE_LITERAL,$abookNewName)) {
                $biNewName = sprintf("{%d}",strlen($abookNewName));
                
                if (!$this->imspSend($biNewName,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                $this->imspRecieve(); //OK+
        }

        //CRLF since last part
        if (!$this->imspSend($abookNewName,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }
        
        
        //Get server response
        $server_response = $this->imspRecieve();                                                                                
        
        switch ($server_response) {
                case "NO":
                        //sorry, can not do it...
                        $this->imspError("Perhaps the addressbook $abookOldName doesn\'t exist or $abookNewName is already taken.");
                        return FALSE;
                        
                case "BAD":
                        //Syntax prob
                        $this->imspError($command_text);
                        return FALSE;
                        
                case "OK":
                        if ($this->log_stream) {
                                $this->write_to_log("Addressbook $abookOldName successfully changed to $abookNewName");
                        }
                        return TRUE;
                
                default:
                        //something unexpected
                        $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                        $this->imspError();
                        return FALSE;
        }               

}


/****************************************************************************************************
*       $entryInfo should be an associative array containing the entry field names.
*        there MUST be a KEY named "name" that contains the name of the entry in the abook.
/****************************************************************************************************/

function addAddress($abook,$entryInfo) {
        
        //$entryInfo must be an array
        if (getType($entryInfo) != "array") {
                $this->exitCode = IMSP_EXIT_BAD_ARGUEMENT;
                $this->imspError("In method ->addAddress() \$entryInfo must be an array.");
                return FALSE;
        }
        
        //First, we should lock the entry if it already exists
        if (!$this->lockABook($abook,$entryInfo["name"])) {
                //Could not obtain a lock on this entry.
                //Error codes would have been dealt with already in the lockABook() call.
                return FALSE;
        }
        
        if (!$this->imspSend("STOREADDRESS ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //First, take care of the name
        $entryName = $entryInfo["name"];
        
        //{} for book name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                $this->imspRecieve(); //+OK
        }
        
        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        
        
        //Do we need {} for entry name as well?
        if (preg_match(IMSP_MUST_USE_LITERAL,$entryName)) {
                $biname=sprintf("{%d}",strlen($entryName));
                
                if (!$this->imspSend($biname,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                $this->imspRecieve(); //+OK
                
                if (!$this->imspSend($entryName,FALSE,FALSE)) {
                        $this->imspError();
                        return FALSE;
                }

        } else {
                
                if (!$this->imspSend("\"$entryName\" ",FALSE,FALSE)) {
                        $this->imspError();
                        return FALSE;
                }

        }

        while (list($key,$value) = each($entryInfo)) {
                //Do not sent the key name "name"       
                if ($key != "name") {                                                                                                           

                        //Protect from extraneous white space
                        $value = trim($value);
                        
                        //For some reason, tabs seem to break this so we should replace them with spaces?
                        $value = preg_replace("/\t/","\n\r",$value);
                        
                        //Check to see if we need {} 
                        if (preg_match(IMSP_MUST_USE_LITERAL,$value)) {
                                $command_text .= $key . sprintf(" {%d}",strlen($value));
                                
                                if (!$this->imspSend($command_text,FALSE,TRUE)) {                                               //send  the octet count at the end
                                        $this->imspError();
                                        return FALSE;
                                }

                                $server_response = $this->imspRecieve();                                                                //OK+
                                $command_text = "";                                                                                                             //Clear the command_text buffer
                                
                                //Check that we DID get a OK+
                                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$server_response)) {
                                        //not expected
                                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                                        $this->imspError($server_response);
                                        return FALSE;
                                }
                                
                                //Send the string of octets and be sure NOT to end with CRLF
                                if (!$this->imspSend($value,FALSE,FALSE)) {
                                        $this->imspError();
                                        return FALSE;
                                }
                                
                        } else {
                                /*If we are here, then we do not need to send a string literal (yet)*/
                                //enclose all values in quotes just in case
                                $value = "\"" . $value . "\"";
                                $command_text .= $key . " " . $value . " ";                                                             
                        }
                }
                        
        } //End while
        
        //Send anything that is left of the command
        if (!$this->imspSend($command_text,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }
        
        //Check on success
        $server_response = $this->imspRecieve();
        
        
        //Decide on the response...
        switch ($server_response) {
                case "NO":
                        //Sorry...can not do it.
                        $this->imspError("Can not add the requested address.");
                        return FALSE;
                        
                case "BAD":
                        //Sorry...did not understand you
                        $this->imspError($command_text);
                        return FALSE;
        }
        
        if ($server_response != "OK") {
                //Cyrus-IMSP server sends a FETCHADDRESS Response here.  Do others?     This was not in the RFC.
                $dummy_array = $this->parseFetchAddressResponse($server_response);
                $server_response = $this->imspRecieve();
                                                                                                                
                
                //Check it again
                switch ($server_response) {
                        case "NO":
                                $this->imspError("Can not add the requested address.");
                                return FALSE;
                        case "BAD":
                                $this->imspError($command_text);
                                return FALSE;
                        case "OK":
                                if ($this->log_stream) {
                                        $this->write_to_log("STOREADDRESS Completed successfully.");
                                }
                                
                                
                                //we were successful...so release the lock on the entry
                                if (!$this->unlockABook($abook,$entryInfo["name"])) {
                                        //could not release lock
                                        $this->exitCode = IMSP_ENTRY_LOCKED;
                                        $this->imspError();
                                        return FALSE;
                                }
                                
                                return TRUE;
                }
        }
}


/****************************************************************************************************
* Deletes an abook entry.  Takes the name of the abook and the name of the entry ($bookEntry)
* Returns true / false
/****************************************************************************************************/
function deleteAddress($abook,$bookEntry) {
        //Start the command
        $this->imspSend("DELETEADDRESS ",TRUE,FALSE);

        //Need {} for book name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
                
                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }       
        
        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        

        //How bout for the entry name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$bookEntry)) {
                $biEntry = sprintf("{%d}",strlen($bookEntry));
                
                if (!$this->imspSend($biEntry,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
        
        } else {
                $bookEntry = $this->quoteSpacedString($bookEntry);
        }

        if (!$this->imspSend($bookEntry,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }
        
        $server_response = $this->imspRecieve();
        
        switch ($server_response) {
                case "NO":
                        //Sorry..can not do it
                        $this->imspError($command_text);
                        return FALSE;
                case "BAD":
                        //Do not know what your talking about
                        $this->imspError($command_text);
                        return FALSE;
                case "OK":
                        //everything is ok!!
                        
                        if ($this->log_stream) {
                                $this->write_to_log("DELETE Completed successfully.");
                        }
        
                        return TRUE;
        }

}

/****************************************************************************************************
* function lockABook - attempts to acquire a semephore on the addressbook entry, $bookEntry in 
* addressbook $abook.  Will return TRUE || $dummy on success.  Will return FALSE on failure.
*****************************************************************************************************/
function lockABook($abook,$bookEntry) {
         

        $this->imspSend("LOCK ADDRESSBOOK ", TRUE, FALSE);
        
        
        //Do we need a string literal?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //What about the entry name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$bookEntry)) {
                $biEntry = sprintf("{%d}", strlen($bookEntry));
                
                if (!$this->imspSend($biEntry,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
                if (!$this->imspSend($bookEntry,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
        
        } else {
                
                $bookEntry = $this->quoteSpacedString($bookEntry);
                
                if (!$this->imspSend("$bookEntry",FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

        }
        
         $server_response = $this->imspRecieve();
        
         do  {
                 switch($server_response) {
                        case "NO":
                                //Could not acquire lock..maybe someone else has it??
                                $this->exitCode = IMSP_ENTRY_LOCKED;
                                $this->imspError();
                                return FALSE;
                        case "BAD":
                                //Syntax problem
                                $this->exitCode = IMSP_SYNTAX_ERROR;
                                $this->imspError();
                                return FALSE;
                 }
        
        
                //Check to see if this is a FETCHADDRESS resonse
                //Do all IMSP implementations return a FETCHADDRESS here?
                $dummy = $this->parseFetchAddressResponse($server_response);
                
                if ($dummy) {
                        //OK
                        $server_response = $this->imspRecieve();
                        $this->write_to_log($server_response);
                        
                }
                
                
                
        } while($server_response != "OK");
        
        //tell the log.
        if ($this->log_stream) {
                $this->write_to_log("LOCK ADDRESSBOOK on $abook $bookEntry OK");
        }
        
        //Return either true of the fetchaddress response if it exists.
        if (!$dummy) {
                return true;
        } else {
                return $dummy;
        }
}


/****************************************************************************************************
* Unlocks a previously locked abook.  Takes the name of the addressbook and the name of the entry.
* Returns True or False on success or failure.
/****************************************************************************************************/
function unlockABook($abook,$bookEntry) {
        
        //Start sending command
        if (!$this->imspSend("UNLOCK ADDRESSBOOK ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        
        //{} for book name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        
        //How bout for entry name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$bookEntry)) {
                $biEntry=sprintf("{%d}",strlen($bookEntry));
                
                if (!$this->imspSend($biEntry,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
                if (!$this->imspSend($bookEntry,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
        
        } else {
                
                $bookEntry = $this->quoteSpacedString($bookEntry);
                
                if (!$this->imspSend("$bookEntry",FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

        }
        
        $response = $this->imspRecieve();
        
        switch ($response) {
                case "NO":
                        //Could not release the lock for some strange reason...maybe we Do not own it?
                        $this->exitCode = IMSP_NO;
                        $this->imspError("Could not release the lock...perhaps we are not the owner of the lock.");
                        return FALSE;
                case "BAD":
                        //Some type of syntax error
                        $this->exitCode = IMSP_SYNTAX_ERROR;
                        $this->imspError("Error releasing lock in UNLOCK command.");
                        return FALSE;
                case "OK":
                        //Tell the log
                if ($this->log_stream) {
                        $this->write_to_log("UNLOCK ADDRESSBOOK on $abook $bookEntry OK");
                }
                
                return TRUE;
        }
}               

/****************************************************************************************************
* Access Control List (ACL)  Methods.
*  
* The following characters are recognized ACL characters: lrwcda
* l - "lookup"  (allows user to see the name and existence of the addressbook)
* r - "read"    (allows searching and retreiving addresses from addressbook)
* w - "write"   (allows creating/editing new addressbook entries - not deleting)
* c - "create"  (allows creating new addressbooks under the current addressbook hierarchy)
* d - "delete"  (may delete entries or entire book)
* a - "admin"   (privledge to set ACL lists for this addressbook - usually only allowed for the owner of the addressbook)
*
*examples:
* "lr" would be read only for that user
* "lrw" would be read/write
****************************************************************************************************/


/****************************************************************************************************
* Sets an Access Control List for an abook.
* takes the abook name, the username ($ident) and a string containing the ACL characters ($acl)
* The ACL string should be a standard ACL type listing of characters such as "lrw" for read/write
* IMSP COMMAND TEXT: SETACL ADDRESSBOOK $abook $ident $acl
/****************************************************************************************************/

function setACL($abook,$ident,$acl) {
        
        //Verify that $acl looks good...
        if (preg_match("/[^" . IMSP_ACL_RIGHTS . "]/",$acl)) {
                //error...acl list contained unrecoginzed options
                $this->exitCode = IMSP_BAD_ARGUMENT;
                $this->imspError("the setACL() method only accepts the following characters in the ACL " . IMSP_ACL_RIGHTS . ".");
                return FALSE;
        }
        
        //Begin sending command
        if (!$this->imspSend("SETACL ADDRESSBOOK ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        
        //{} for book name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if(!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
                
                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //{} for ident?
        if (preg_match(IMSP_MUST_USE_LITERAL,$ident)) {
                $biIdent = sprintf("{%d}",strlen($ident));
                
                if (!$this->imspSend($biIdent,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
                
                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {
                        $this->imspError();
                        return FALSE;
                }
        
        }

        if (!$this->imspSend("$ident ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //now finish up with the actual acl...
        if (!$this->imspSend($acl,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }

        $response = $this->imspRecieve();
        
        switch ($response) {
                case "NO":
                        //Could not set ACL
                        $this->imspError("$ident ACL could not be set for addressbook $abook");
                        return FALSE;
                case "BAD":
                        //Bad syntax
                        $this->imspError();
                        return FALSE;
                case "OK":
                        //COOL!!
                        if ($this->log_stream) {
                                $this->write_to_log("ACL set for $ident on $abook");
                        }
                        return TRUE;
                        
                default:
                        //Do not know why we would make it down here, so return FALSE for now
                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                        $this->imspError();
                        return FALSE;
        }
        


}


/****************************************************************************************************
* Retrieves an addressbooks ACL. 
* This function returns an associatve array containing the name of the user as the key and the
* ACL string as the value so you would get an array such as this:
*
* $result['jsmith'] = "lrw"
* $result['jdoe'] = "r"
* IMSP SYNTAX: GETACL ADDRESSBOOK $abook
****************************************************************************************************/


function getACL($abook) {
        
        if (!$this->imspSend("GETACL ADDRESSBOOK ",TRUE,FALSE)) {
                //ERROR
        }

        //{} for book name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biName = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biName,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
                
                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
        }

        if (!$this->imspSend($abook,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }

        //Get results.
        $response = $this->imspRecieve();
        
        switch($response) {
                case "NO":
                        //Could not complete?
                        $this->imspError("Could not retrieve ACL. Perhaps the addressbook does not exist?");
                        return FALSE;
                case "BAD":
                        //Do not know what you said!
                        $this->imspError();
                        return FALSE;
        }
        
        //If we are here, we need to recieve the * ACL Responses
                

                do {    
                        /* Get an array of responses. 
                         * The [3] element should be the addressbook name
                         * [4] and [5] will be user/group name and permissions etc...
                         */
                        
                        //the book name might be a {}
                        if (preg_match(IMSP_OCTET_COUNT,$response,$tempArray)) {
                                $byteCount = $tempArray[2];//???
                                $data = $this->recieveStringLiteral($byteCount);
                                //Get the rest
                                $response = $this->imspRecieve();
                        }

                        $acl = split(" ",$response);
                        
                        //push the array if there was a {}
                        if ($data) {
                                array_unshift($acl," "," "," "," ");
                        }
                        
                        for ($i = 4 ; $i < count($acl) ; $i += 2) {
                                $results[$acl[$i]] = $acl[$i+1];
                        }
                        
                        $response = $this->imspRecieve();
                        
                
                } while (preg_match(IMSP_ACL_RESPONSE,$response));
                
                //Hopefully we can recieve an OK response here
                
                if ($response != "OK") {
                        //some weird problem
                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
                if ($this->log_stream) {
                        $this->write_to_log("GETACL on $abook completed.");
                }
                
                return $results;
                
}


/****************************************************************************************************
* Deletes an ACL entry for a abook.  Takes the abook name and the username whose ACL should be deleted.
* IMSP SYNTAX: DELETEACL ADDRESSBOOK $abook $ident
****************************************************************************************************/
function deleteACL($abook,$ident) {

        if (!$this->imspSend("DELETEACL ADDRESSBOOK ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }
        
        //{} FOR BOOK NAME?
        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if (!$this->imspSend("$abook ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //{} for ident name?
        if (preg_match(IMSP_MUST_USE_LITERAL,$ident)) {
                $biIdent = sprintf("{%d}",strlen($ident));
                
                if (!$this->imspSend($biIdent,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
                if (!$this->imspSend($ident,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

        } else {
                
                if (!$this->imspSend("\"$ident\"",FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }
        }

        //Get results
        $server_response = $this->imspRecieve();
        
        switch($response) {
                case "NO":
                        //could not complete
                        $this->imspError("Could not delete the ACL for $ident on addressbook $abook.");
                        return FALSE;
                case "BAD":
                        //Do not understand!
                        $this->imspError();
                        return FALSE;
                case "OK":
                        //COOL!
                        if ($this->log_stream) {
                                $this->write_to_log("DELETED ACL for $ident on $abook");
                        }
                        return TRUE;
                
                default:
                        //Do not know why we would be here?
                        $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                        $this->imspError();
                        return FALSE;
        }
        
}


/****************************************************************************************************
 * Returns an ACL string containing the rights for the currently logged in user for the addressbook
 * passed in $abook.  Returns FALSE on failure.
 *IMSP SYNTAX: MYRIGHTS ADDRESSBOOK $abook
/****************************************************************************************************/
function myRights($abook) {

        if (!$this->imspSend("MYRIGHTS ADDRESSBOOK ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        if (preg_match(IMSP_MUST_USE_LITERAL,$abook)) {
                $biBook = sprintf("{%d}",strlen($abook));
                
                if (!$this->imspSend($biBook,FALSE,TRUE)) {
                        $this->imspError();     
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }

        }

        if (!$this->imspSend($abook,FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }

        //Get results.  
        $server_response = $this->imspRecieve();
        
        switch($server_response) {
                case "NO":
                        //could not complete
                        $this->imspError("Could not retrieve the ACL for the current user ($this->user)");
                        return FALSE;
                case "BAD":
                        //Do not understand!
                        $this->imspError();
                        return FALSE;
        }
        
        if (!preg_match(IMSP_MYRIGHTS_RESPONSE,$server_response)) {
                $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                $this->imspError();
                return FALSE;
        }       
        

        //Did the response have a {} for the abook name?
        if (preg_match(IMSP_OCTET_COUNT,$server_response,$tempArray)) {
                $dataSize = $tempArray[2];
                $data = $this->recieveStringLiteral($dataSize);
                //Get the rest
                $server_response = $this->imspRecieve();
        }

        $temp = split(" ",$server_response);                    
        
        //Push the array if we had a {}
        if ($data) {
                array_unshift($temp," "," "," "," ");
        }
        
        $acl = $temp[4];        
        
        //Get the OK response
        $server_response = $this->imspRecieve();
        
        //Check for OK?
        if ($server_response != "OK") {
                $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                $this->imspError();
                return FALSE;
        } else {
                if ($this->log_stream) {
                        $this->write_to_log("MYRIGHTS on $abook completed.");
                }
                return $acl;
        }       
        
}       
/****************************************************************************************************
* Utility Functions to support the class
* These would be considered "Private" but PHP has no support for this distinction.
/****************************************************************************************************/

function getXMLAddressbookEntry($results) {
        
        $xml = "<addressbookentry>\n";
        
        //Protects against FALSE being returned.  In this case, there are no results, so return an
        //empty XML instance of <addressbookentry>
        if (!$results) {
                $results = array ();
        }
        
        while (list($key,$value) = each($results)) {
                $xml .= "\t<" . $key .  ">" . $value . "</" . $key . ">\n";
        }
        
        $xml .= "</addressbookentry>\n";

        return $xml;


}


function parseFetchAddressResponse($server_response) {
        
        /*
         * Expects a FETCHADDRESS response to be passed.  Parses it out into an associative array
         * with name-value pairs from the address book entry
         *
         *TODO: FIgure out why we Can not escape the first quotes!
         */
        
        if(!preg_match(IMSP_FETCHADDRESS_RESPONSE,$server_response)) {
                
                if($this->log_stream) {
                        $this->write_to_log("[ERROR] Did not recieve a FETCHADDRESS response from server.");
                }
                
                //Try to decide what the response was here.
                $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                return FALSE;
        } 
        
        /****************************************************************************************************
         *Parse out the server response string
        *
        * After choping off the server command response tags and split()'ing the server_response string using
        * a " " as the delimiter, the $parts array contains the chunks of the server returned data.
        *
        * The predifined "name" field starts in $parts[1].  The server should return any single item of data
        * that contains spaces within it as a double quoted string.  So we can interpret the existence of a 
        * double quote at the beginning of a chunk to mean that the next chunk(s) are to be considered part of 
        * the same value.  A double quote at the end of a chunk signifies the end of that value and the chunk
        * following that can be interpreted as a key name.
        *
        * We also need to watch for the server returning a {} response for the value of the key as well.
        *****************************************************************************************************/
        
        //Ooopss..check to see if the abook name is itself a {}.
        if (preg_match("/(^\* FETCHADDRESS )({)([0-9]{1,})(\}$)/",$server_response,$tempArray)) {
                
                $abook = $this->recieveStringLiteral($tempArray[3]);
                $chopped_response = trim($this->imspRecieve());
        
        } else {
                
                //Take off the stuff from the beginning of the response we Do not need.
                $chopped_response = trim(preg_replace(IMSP_FETCHADDRESS_RESPONSE,"",$server_response));                                 
        
        }

        $parts = split(" ",$chopped_response);
        //If abook was sent as a {} then we must "push" a blank
        //value to the start of this array so the rest of the routine
        //will work with the correct indexes!!
        if ($abook) {
                array_unshift($parts," ");
        }

        $numOfParts = count($parts);
        
        //This is the name - possibly only the first part.
        $name = $parts[1];                                                                                                                                                                      
        
        //Get the first char of the name string                 
        $firstChar = substr($name,0,1);                                                                                                                                         
                                
        /* Check to see if the first char of the name string is a double quote so */
        /* we know if we have to extract more of the name in the following chunks */
                                 
        if($firstChar == "\"") {
                
                for($i = 2 ; $i < $numOfParts; $i++){
                        
                        $name .=  " " . $parts[$i];
                        $lastChar = substr($parts[$i],strlen($parts[$i]) - 1,1);
                        
                        if($lastChar == "\""){
                                $nextKey = $i + 1;
                                break;
                        }
                }
        // could also be a string literal.... veg
        } elseif (preg_match('/\{(\d+)\}/',$name,$matches)) {
                $nameLen = $matches[1];
                $name = $this->recieveStringLiteral($nameLen);
                $response=$this->imspRecieve();
                $parts = split(" ",$response);
                $numOfParts = count($parts);
                $nextKey = 0;
        } else {
        
                /* If only one chunk for 'name' then we just have to point to the */
                /* next chunk in the array...which will hopefully be '2'.         */
                $nextKey = 2;
        }
        
        
        //remove any quotes sent to us by the server
        if (substr($name,0,1)=="\"") {
                $name = substr($name,1,strlen($name));
        }

        if (substr($name,strlen($name)-1,1) == "\"") {
                $name = substr($name,0,strlen($name) - 1);
        }

        //escape the name here?
        if ($this->escape_special_characters) {
                $name = htmlspecialchars($name);
        }
                                                
        $lastChar="";                                                                                                                                                           
                                
        //Fill in the name into the array we will return.                                                                                                               
        $entry["name"] = $name;                                                                                                                                                                 
                                
        //Start parsing the rest of the response.
        for($i = $nextKey ; $i < $numOfParts ; $i += 2) {
                $key = $parts[$i];

                //Check for a literal string response {}.
                if(preg_match("/(^{)([0-9]{1,})(\}$)/",$parts[$i+1],$tempArray)) {                                                                                      
                        $dataSize = $tempArray[2];                                                                                                              
                        $server_data = $this->recieveStringLiteral($dataSize);                                                                                  
                        $entry[$key] = $server_data;
                        
                        //Escape here?
                        if ($this->escape_special_characters){
                                $entry[$key] = htmlspecialchars($entry[$key]);
                        }
                                                
                        //Read any remaining data from the stream and reset the counter variables
                        //so the loop will continue correctly. Note we set $i to -2 because it will
                        //be incremented by 2 before the loop will run again.
                        
                        $parts = $this->getServerResponseChunks();
                        $i= -2;
                        $numOfParts = count($parts);
                        
                } else {//not a string literal response
                                        
                        $entry[$key] = $parts[$i + 1];
                
                        /* Check to see if the value started with a double quote.     
                         * This signifies that the value continues to the next element
                         * We also need to check if the last char is a quote to make sure we REALLY have
                         * to check the next elements for a closing quote!!!
                         */
                                         
                        if ((substr($parts[$i+1],0,1) == "\"") &&(substr($parts[$i+1],strlen($parts[$i+1]) - 1,1) != "\"")){
                                                
                                do{
                                        $nextElement = $parts[$i+2];
                                        
                                        
                                        //Was this element the last one? We have to check this BEFORE we
                                        //escape the special characters!!
                                        $lastChar = substr($nextElement,strlen($nextElement) - 1,1);
                                        $entry[$key] .= " " . $nextElement;     
                                
                                        //NOW, we can check the lastChar since we got it before we f'd with the string!
                                        if ($lastChar=="\"") {
                                                $done = TRUE;
                                                $i++;
                                        } else {                                                        
                                                //Check to see if the next element is the last one
                                                //If so, the do loop will terminate.
                                                $done = FALSE;
                                                $lastChar = substr($parts[$i+3],strlen($parts[$i+3]) - 1,1);
                                                $i++;
                                        }
                                                                                
                                } while ($lastChar != "\"");
                                        
                                //Do we need to add the final element, or was there only two total?
                                if(!$done){
                                        $nextElement = $parts[$i+2];
                                        $entry[$key] .= " " . $nextElement;
                                        $i++;
                                }
                                
                                //remove the quotes sent back to us from the server
                                if (substr($entry[$key],0,1)=="\"") {
                                        $entry[$key] = substr($entry[$key],1,strlen($entry[$key])-2);
                                }
                                
                                
                                if (substr($entry[$key],strlen($entry[$key]) - 1,1) == "\"") {
                                        $entry[$key] = substr($entry[$key],0,strlen($entry[$key]) - 2);
                                        
                                }
                                
                                //Escape here?
                                        if ($this->escape_special_characters) {
                                                $nextElement = htmlspecialchars($nextElement);
                                        }


                        }
                                         
                }//end of else clause (not literal string)
        }//end of for i loop                            
        return $entry;
}


} //End IMSPABook Class

?>
