<?php
 /******************************************************************************************************
 * Project: PHP IMSP CLIENT LIBRARY  
 * FILE:                cIMSPStream.PHP
 * PROVIDES:    Basic IMSP connectivity and defines the CIMSPStream object.
 *
 * Copyright 2002 Michael Rubinsky <mike@theupstairsroom.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ********************************************************************************************************/

//Constant Definitions
define("IMSP_CRLF","\r\n");
define("IMSP_RESPONSE_TAG_LENGTH",5);
define("IMSP_DEFAULT_RESPONSE_LENGTH",512);
define("IMSP_XML_HEAD","<?xml version=\"1.0\"?>\n");

//These define regExp that should match the respective server response strings
define ("IMSP_CONNECTION_OK","/^\* OK/");
define ("IMSP_LOGIN_OK","/OK User/");
define ("IMSP_COMMAND_CONTINUATION_RESPONSE","/^\+/");
define ("IMSP_CAPABILITY_RESPONSE","/^\* CAPABILITY/");


//Exit code values (these define the text that will be logged or displayed in the case of errors)
define ("IMSP_EXIT_LOGIN_FAILED","Login to host failed");
define ("IMSP_CONNECTION_FAILURE","Connection failure");
define ("IMSP_NOT_EXPECTED_RESPONSE", "Did not recieve the expected response from the server.");
define ("IMSP_EXIT_UNEXPECTED_RESPONSE", "Did not recieve the expected response from the server.");             //For compatability
define ("IMSP_SYNTAX_ERROR","The server did not understand your request.");
define ("IMSP_BAD_SYNTAX","The server did not understand your request.");                                                               //For backwards compatability.
define ("IMSP_NO","Server returned a NO response");             
define ("IMSP_EXIT_BAD_ARGUMENT","The wrong type of arguement was passed to this function");            
define ("IMSP_NO_CONTINUATION_RESPONSE","Did not receive expexted command continuation response from server.");
                                                
class IMSPStream {
        /****************************************************************************************
        * Public Property Declarations
        ****************************************************************************************/
        var $imsp_server = "158.223.1.3";                                  //Defaults to "localhost" for the IMSP server host      
        var $imsp_port = "406";                                          //...on the default port
        var $user;                                                       //Property for the username
        var $pass;                                                       //Holds the user's password
        var $auth_method = "PLAIN";                                      //Set this to the type of Authentication Method to be used.
        var $exitCode;                                                   //Routines may place exit codes into this property for checking by the client application
        var $log_name = "imspstream.log";                                 //The name of the log to write to if logging enabled.
        var $logging = TRUE;                                              //Set this to "TRUE" to enable logging.
        var $log_level = 2;                                               // 1 = Normal user level logging 
                                                                          // 2 = Include client debugging info (response strings etc..)
        var $seperator = ".";                                             //Heirarchy seperator (STILL NEED THIS??)
        var $showImspErrors = TRUE;                                       //Determines if IMSP errors are passed thru to the client.
        var $debugging = TRUE;                                            //If true, will turn on PHP's error messages.
        var $escape_special_characters = TRUE;                            //Set to true to escape special HTML/XML characters.
        
        /***************************************************************************************
        * "Private" or "Friend" Declarations
        ****************************************************************************************/
        var $commandPrefix = "A";                                         //Holds the current alphabetic portion of command tag   
        var $commandCount = 1;                                            //Holds the current numeric portion of the command tag
        var $tag = "";                                                   //Holds the most current command tag
        var $stream;                                                    //reference to the file pointer for the IMSP stream
        var $log_stream;                                                 //file stream to log file;
        var $connectionID;                                               //Holds uniques identifier (for logging)
        
        //Holds the last commandtag used (used to help prevent reading the wrong response if the full response was not recieved
        //must set this to some value here to prevent the logon responses from causing a failure!
        var $lastCommandTag = "undefined";                                                              

function IMSPStream($local_imsp_server="", $local_port="") {
        
        // Check for passed in values for server and port number
        if ($local_imsp_server) {
                $this->imsp_server = $local_imsp_server;
        }
        
        if($local_port) {
                $this->imsp_port = $local_port;
        }
        
        $this->connectionID = time();                    //get a somewhat unique identifier...good enough for logging purposes.
        
        //See if we need to turn on error messages or not.
        if($this->debugging) {
                ini_set("display_errors","1");          
        } else {
                ini_set("display_errors","0");
        }                       
                                
 }

function login($username,$pass) {
        
        //base64 encode the password in case this
        //object is saved to a session cache.
        //$this->user = base64_encode($username);
        $this->user = $username;
        $this->pass = base64_encode($pass);
        
        
        if ($this->logging) {
                $this->open_log($this->log_name);       
                $this->write_to_log("Starting IMSP Session");
        }
        
        
        // Now try to open the IMSP connection
        if (!$this->imspOpen()) {
                $this->imspError();
                return FALSE;
        }
         
         if ($this->log_stream) {
                $this->write_to_log("Connection OK");
         }

 
        /*Everything is ok...so try to login to the server
         *For now, we do things simply, using LOGIN command
         *
         *in release version, we will want to support SASL logins
         */
         Switch ($this->auth_method) {
                case "PLAIN":
                        if (!$this->imspPlainTextLogin()) {
                                //Login failed
        
                                if($this->exitCode = "") {
                                        $this->exit_code = IMSP_EXIT_LOGIN_FAILED;
                                }
                                //$this->imspError();
                                return FALSE;
                        }
                        
                        break;

                case "CRAM-MD5":
                        if (!$this->cramMD5Auth()) {
                                //login failed
                                
                                if($this->exitCode = "") {
                                        $this->exit_code = IMSP_EXIT_LOGIN_FAILED;
                                }
                        
                                return FALSE; 
                        }

                        break;


                default:
                        //must specify an auth method 
                        $this->exit_code = IMSP_EXIT_LOGIN_FAILED;
                        $this->imspError("Must specify an authentication method in the auth_method property.");
                        return FALSE;
         }
         
        //OK. We are logged in to the server
        
        if ($this->log_stream) {
                $this->write_to_log("LOGIN for user $this->user successful.");
         }
        return TRUE;

}
/****************************************************************************************
 *logout() - logs out of the server and closes the IMSP stream
 ****************************************************************************************/
function logout() {
        //Disconnect from the server
        
        if($this->log_stream){
                $this->write_to_log("Closing Connection.");
        }
        
        
                $command_string = "LOGOUT";                             //Build the command to send to the server
        
                if (!$this->imspSend($command_string)) {
                        $this->imspError();
                        return FALSE;
                } else {
                        //Should we test here for the BYE response? Is there a need for this?
                        
                        fclose($this->stream);
                        
                        if ($this->log_stream) {
                                $this->close_log();
                        }
                        
                        return TRUE;
                }
                
        
}


/**************************************************************************************
 *Capability command
 *returns the raw capability string from the server
 *************************************************************************************/
 function capability() {
         
        $command_string = "CAPABILITY";
        if (!$this->imspSend($command_string)) {
                        $this->imspError();
                        return FALSE;
        } else {
                        $server_response = $this->imspRecieve();

                        if (preg_match(IMSP_CAPABILITY_RESPONSE, $server_response)) {
                                $capability = preg_replace(IMSP_CAPABILITY_RESPONSE,"",$server_response);
                                $server_response = $this->imspRecieve(); //OK
                                if (!$server_response == "OK") {
                                        $this->exitCode = IMSP_NOT_EXPECTED_RESPONSE;
                                        $this->imspError($server_response);
                                } else {
                                        if ($this->log_stream) {
                                                $this->write_to_log("CAPABILITY completed OK");
                                                return $capability;
                                        }
                                }

                        }
        }


 }


/**************************************************************************************
 * IMSP connectivity functions
 *************************************************************************************/

function imspOpen() {
        // Now try to open the IMSP connection
        $fp = fsockopen($this->imsp_server,$this->imsp_port);
        
        
        // Check for failure
        if(!$fp) {
                $this->exit_code = IMSP_CONNECTION_FAILED;
                $this->imspError();
                return FALSE;   
        }
        
        //save the file pointer
        $this->stream = $fp;

        //Get The Server Response
        $server_response = $this->imspRecieve();
                
                        
        //Check that it is what was expected
        if(!preg_match(IMSP_CONNECTION_OK,$server_response)) {
                $this->exit_code = IMSP_NOT_EXPECTED_RESONSE;
                $this->imspError();
                fclose($fp);
                return FALSE; 
                
        }

        return TRUE;
        
}


function imspSend($commandText, $includeTag = TRUE, $sendCRLF=TRUE){
        
        if(!$this->stream){
                //no connection!!
                $this->exit_code = IMSP_CONNECTION_FAILURE;
                return FALSE;
        }
        
        if ($includeTag) {
                $this->tag = $this->getNextCommandTag();
                $command_text = "$this->tag ";
                //$this->lastCommandTag = $this->tag;
        }
        
        $command_text .= $commandText;
        
        
        if ($sendCRLF) {        
                $command_text .= IMSP_CRLF;
        }
        
        
        if ($this->log_stream && $this->log_level > 1) {
                $this->write_to_log("[CLIENT COMMAND] $command_text");
        }
        
        if (!fputs($this->stream,$command_text)) {
                //something wrong with connection?
                $this->exit_code = IMSP_CONNECTION_FAILURE;
                return FALSE;
                
        } else {
        
                return TRUE;
       }
        
}

function imspRecieve() {
        //Recieves a single CRLF terminated server status response from the server.
        
        if (!$this->stream){
                //no connection!!
                $this->exit_code = IMSP_CONNECTION_FAILURE;
                return FALSE;
        }
        
        $server_response = trim(fgets($this->stream,IMSP_DEFAULT_RESPONSE_LENGTH));
        
        if ($this->log_stream && $this->log_level > 1) {
                $this->write_to_log("[SERVER RESPONSE] $server_response");
        }
        
        //should we see if the response is simply an OK, BAD or NO in response to an tagged command?
        
        //Parse out the response
        
        //First make sure that this is not for a previous command.
        //If it is, it means we did not read all the server responses from the last command...read them now, 
        //but throw an error.
        while (preg_match("/^" . $this->lastCommandTag ."/",$server_response)) {
                $server_response = trim(fgets($this->stream,IMSP_DEFAULT_RESPONSE_LENGTH));
                $this->exitCode = IMSP_EXIT_UNEXPECTED_RESPONSE;
                $this->imspError($server_response);
        }
        
        $currentTag = $this->tag;
        if (preg_match("/^" . $currentTag . " NO/", $server_response)) {
                //Could not perform action.
                $this->exitCode = IMSP_NO;
                return "NO";

        }
        
        if (preg_match("/^" . $currentTag . " BAD/",$server_response)) {
                //Bad syntax or entry names
                $this->exitCode = IMSP_SYNTAX_ERROR;
                return "BAD";
        }
        
        if (preg_match("/^" . $currentTag . " OK/",$server_response)) {
                return "OK";
        }

        //If it was not a "NO", "BAD" or "OK" response, then it is up to the 
        //calling function to decide what to do with it 
        return $server_response;
}

function getServerResponseChunks(){
        //Retrieves a CRLF terminated response from the server and splits it into an array delimeted by a <space> and returns array
        $server_response = trim(fgets($this->stream,IMSP_DEFAULT_RESPONSE_LENGTH));
        $chunks = split(" ",$server_response);
        return $chunks;
}



function recieveStringLiteral($length){
        $temp = trim(fread($this->stream,$length));

        if (($this->log_stream) && ($this->log_level > 1)) {
                $this->write_to_log("[SERVER RESPONSE] $temp");
        }

        return $temp;

}

/**************************************************************************************
 *Authentication Routines
 *************************************************************************************/
 function imspPlainTextLogin() {
        $user= $this->user;
        $pass = base64_decode($this->pass);
         
         //make sure any logging is at level 1 only to ensure no password information
        //will be sent to the log during the imspSend() call.
        $logLevel = $this->log_level;
        $this->log_level = 1;

        //start the command...
        if (!$this->imspSend("LOGIN ",TRUE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //username as a {}?
        if (preg_match(IMSP_MUST_USE_LITERAL,$user)) {
                $biUser = sprintf("{%d}",strlen($user));

                if (!$this->imspSend($biUser,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
        }
        
        if (!$this->imspSend("$user ",FALSE,FALSE)) {
                $this->imspError();
                return FALSE;
        }

        //Pass as {}?
        if (preg_match(IMSP_MUST_USE_LITERAL,$pass)) {
                $biPass = sprintf("{%d}",strlen($pass));

                if (!$this->imspSend($biPass,FALSE,TRUE)) {
                        $this->imspError();
                        return FALSE;
                }

                if (!preg_match(IMSP_COMMAND_CONTINUATION_RESPONSE,$this->imspRecieve())) {     //+OK
                        $this->exitCode = IMSP_NO_CONTINUATION_RESPONSE;
                        $this->imspError();
                        return FALSE;
                }
                
        }
        
        if (!$this->imspSend("$pass",FALSE,TRUE)) {
                $this->imspError();
                return FALSE;
        }
        
        //restore original logging level
        $this->log_level = $logLevel;
        
        //Get the response
        $server_response = $this->imspRecieve();
        
        if ($server_response!="OK") {
                //login failed
                $this->exitCode = IMSP_LOGIN_FAILED;
                return FALSE;
        }
        
                return TRUE;
 }
 
 function cramMD5Auth() {
        //$user = base64_decode($this->user);
        $user = $this->user;
        $pass = base64_decode($this->pass);
        
        //Set up the command and send it...
        $command_string = "AUTHENTICATE CRAM-MD5";
        
        if (!$this->imspSend($command_string)) {
                $this->exitCode = IMSP_CONNECTION_FAILURE;
                return FALSE;
        }

        //Get response and decode it...note that we remove the 1st 2 characters
        //from the response to get rid of the "+" continuation character and the space
        //that is sent as part of the CRAM-MD5 response (at least on cyrus-imspd)
        $server_response = $this->imspRecieve();
        $server_response = base64_decode(trim(substr($server_response,2)));
        
        //Build and base65 encode the response to the challange.
        $response_to_send = $user . " " . $this->hmac($pass,$server_response);
        $command_string = base64_encode($response_to_send);
        
        //Send the response
        if (!$this->imspSend($command_string,FALSE)) {
                $this->exitCode = IMSP_CONNECTION_FAILURE;
                return FALSE;
        }

        //See if we are OK!
        $server_response =  $this->imspRecieve();
        
        if ($server_response!="OK") {
                //login failed
                $this->exitCode = IMSP_LOGIN_FAILED;
                return FALSE;
        } else {
        
                return TRUE;
        }
                
 }
 
 
/**************************************************************************************
 * Utility Functions
 *************************************************************************************/
function getNextCommandTag() {
                
                $newtag = $this->commandPrefix;
                
                if ($this->commandCount < 10) {
                        $newtag .= "000" . (string) $this->commandCount;
                } elseif ($this->commandCount < 100){
                        $newtag .= "00" . (string) $this->commandCount;
                } elseif ($this->commandCount < 999) {
                        $newtag .= "0" . (string) $this->commandCount;
                }
                
                //increment it for the next command
                $this->commandCount++;
                $this->lastCommandTag = ($this->tag) ? $this->tag : "undefined";
                return $newtag;
}

function quoteSpacedString($string) {
        
        if (preg_match("/ /",$string)) {
                return "\"" . $string . "\"";
        } else {
                return $string;
        }
}

function imspError($server_text = "") {

        if ($this->log_stream) {
                $this->write_to_log("[ERROR]$this->exitCode");
        }
        
        if ($this->showImspErrors) {
                echo("<B>There was an error with the IMSP service.<BR>\n");
                echo("The error code is: $this->exitCode </B><BR>\n");
                
                if($server_text != "") {
                        echo("The following information is also available: $server_text <BR>\n");
                }
        }
        
        //reset the exitCode
        $this->exitCode = "";
        
}





/******************************************************************************************
 *Logging functions
 *****************************************************************************************/
 
function open_log($logfile){
        
        $fp = @fopen($logfile,"a");
        if (!$fp) {
                return FALSE;
        } else {
                $this->log_stream = $fp;
                return TRUE;
         }
}
        
function write_to_log($message) {
        if ($this->logging) {
                
                if (!$this->log_stream) {
                        $this->open_log($this->log_name);
                }

                @fputs($this->log_stream,date("r") . ":[$this->connectionID][$this->user] $message\n");
        }
}

function close_log(){
        @fclose($this->log_stream);
}

function hmac ($key, $data)
                            {
                               // RFC 2104 HMAC implementation for php.
                               // Creates an md5 HMAC.
                               // Eliminates the need to install mhash to compute a HMAC
                             

                               $b = 64; // byte length for md5
                               if (strlen($key) > $b) {
                                   $key = pack("H*",md5($key));
                               }
                               $key  = str_pad($key, $b, chr(0x00));
                               $ipad = str_pad('', $b, chr(0x36));
                               $opad = str_pad('', $b, chr(0x5c));
                               $k_ipad = $key ^ $ipad ;
                               $k_opad = $key ^ $opad;

                               return md5($k_opad  . pack("H*",md5($k_ipad . $data)));
                            }
}//End of Class Definition

?>
